/* YOODA : Yet another Object Oriented Database
   Copyright (C) 1994  Eric Abecassis

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Library General Public
License as published by the Free Software Foundation; either
version 2 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Library General Public License for more details.

You should have received a copy of the GNU Library General Public
License along with this library; if not, write to the Free
Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. */


#pragma implementation "monitor_svc.hxx"


//#define TRACE

#include <iostream.h>
#include <string.h>
#include <signal.h>
#ifdef __hpux__
extern "C" {
#endif
#include <sys/mman.h>
#ifdef __hpux__
}
#endif
  

#ifndef __task_HXX
#include "task.hxx"
#endif

#ifndef __yooexcept_HXX
#include "yooexcept.hxx"
#endif

#ifndef page_HXX
#include "page.hxx"
#endif

#ifndef dbvue_HXX
#include "dbvue.hxx"
#endif

#ifndef delta_HXX
#include "delta.hxx"
#endif

#ifndef dbserver_HXX
#include "dbserver.hxx"
#endif

#ifndef __var_H
#include "var.h"
#endif

#ifndef __monitor_HXX
#include "monitor.hxx"
#endif

#ifndef __monitor_svc_HXX
#include "monitor_svc.hxx"
#endif


#include "callbackserver_clnt.hxx"
#include "dbvue_svc.hxx"

ServerTask **YOO_Task ;
ListPtr<RemoteDBVue> YOO_ListVues ;

//#define DEBUG

#ifdef DEBUG
static int count_read = 0 ;
#endif


/*= EXTERNAL FUNCTION =======================================================*/
int Terminate() 
/*=
[DESCRIPTION :]
  (IN)  
  (OUT) 
[DIAGNOSTIC :]
[CALL :]
=*/
{
  yoo_debug("YOODA Server shutdown, please wait... ") ;
  delete YOO_DBServer ;
  exit(1) ;
  return 1 ;
}


/*= EXTERNAL FUNCTION =======================================================*/
void RemoteDBVue::connect (char* usid, int pid, char* cbName, int delta)
/*=
[DESCRIPTION :]
  (IN)  
  (OUT) 
[DIAGNOSTIC :]
[CALL :]
=*/
{ 
  DBG(yoo_debug("-> connect", this->_vue));

  if (delta) {
    this->_deltavue = Delta::createVue() ;
  }
  this->_usid = strdup(usid) ;
  this->_pid = pid ;
  this->_callback.import(cbName) ;    
  this->setRunning() ;

  DBG(yoo_debug("<- connect", this->_vue));
}


/*= EXTERNAL FUNCTION =======================================================*/
void RemoteDBVue::disconnect ()
/*=
[DESCRIPTION :]
  (IN)  
  (OUT) 
[DIAGNOSTIC :]
[CALL :]
=*/
{ 
  DBG(yoo_debug("-> disconnect", this->_vue));

  this->abort(FALSE) ;
  this->_callback.close() ;
  this->setIdle() ;

  DBG(yoo_debug("<- disconnect", this->_vue));
}


/*= EXTERNAL FUNCTION =======================================================*/
void RemoteDBVue::upgrade(ListVal<PID>& in, ListVal<PID>& eject)
/*=
[DESCRIPTION :]
  (IN)  
  (OUT) 
[DIAGNOSTIC :]
[CALL :]
=*/
{
  if (_deltavue == NULL) return ;
  IterListVal<PID> iter(in) ;
  for (; iter.more(); iter.next()) {
    if (this->_deltavue->delta(iter.get())) {
      eject += iter.get() ;
    }
  }
  delete this->_deltavue ;
  this->_deltavue = Delta::createVue() ;
}


/*= EXTERNAL FUNCTION =======================================================*/
void RemoteDBVue::release(int delta, ListVal<PID>& keep, 
			  ListVal<PID>& keep_out, 
			  ListVal<PID>& eject_out)
/*=
[DESCRIPTION :]
  (IN)  
  (OUT) 
[DIAGNOSTIC :]
[CALL :]
=*/
{
  DBG(yoo_debug("-> release", this->_vue)) ;

  YOO_DBServer->releaseLocks(this->_vue, this->_locks) ;
  IterListVal<PID> iter(keep) ;
  while (iter.more()) {
    if ((this->_deltavue != NULL) && (this->_deltavue->delta(iter.get()))) {
      eject_out += iter.get() ;
      iter.remove() ;
    } else {
      iter.next() ;
    }
  }
  YOO_DBServer->keepLocks(this->_vue, keep, keep_out, eject_out) ;

  this->_locks.clear() ;
  for (IterListVal<PID> iter2(keep_out); iter2.more(); iter2.next()) {
    this->_locks += iter2.get() ;
  }

  if (_deltavue != NULL) { delete this->_deltavue ; _deltavue = NULL ; }
  if (delta) { this->_deltavue = Delta::createVue() ; }

  DBG(yoo_debug("<- release", this->_vue)) ;
}



/*= EXTERNAL FUNCTION =======================================================*/
int RemoteDBVue::sync()
/*=
[DESCRIPTION :]
  (IN)  
  (OUT) 
[DIAGNOSTIC :]
[CALL :]
=*/
{
  DBG(yoo_debug("-> sync", this->_vue)) ;
  return this->ServerDBVue::sync() ;
  DBG(yoo_debug("<- sync", this->_vue)) ;
}


/*= EXTERNAL FUNCTION =======================================================*/
void RemoteDBVue::abort(int delta)
/*=
[DESCRIPTION :]
  (IN)  
  (OUT) 
[DIAGNOSTIC :]
[CALL :]
=*/
{
  DBG(yoo_debug("-> abort", this->_vue)) ;
 
  this->ServerDBVue::abort(delta) ;
  YOO_DBServer->releaseLocks(this->_vue, this->_locks) ;
  this->_locks.clear() ;
  if (_deltavue != NULL) { delete this->_deltavue ; _deltavue = NULL ;}
  if (delta) { this->_deltavue = Delta::createVue() ; }

  DBG(yoo_debug("<- abort", this->_vue)) ;
}


/*= EXTERNAL FUNCTION =======================================================*/
int RemoteDBVue::secondphase() 
/*=
[DESCRIPTION :]
  (IN)  
  (OUT) 
[DIAGNOSTIC :]
[CALL :]
=*/
{
  DBG(yoo_debug("-> secondphase", this->_vue)) ;
  return this->ServerDBVue::secondphase() ;
  DBG(yoo_debug("<- secondphase", this->_vue)) ;
}


/*= EXTERNAL FUNCTION =======================================================*/
void RemoteDBVue::write(const Page& page, const PID& pid)
/*=
[DESCRIPTION :]
  (IN)  
  (OUT) 
[DIAGNOSTIC :]
[CALL :]
=*/
{
  cerr << "." ;
  this->ServerDBVue::write(page, pid) ;
}


/*= EXTERNAL FUNCTION =======================================================*/
void RemoteDBVue::freepage(const PID& pid)
/*=
[DESCRIPTION :]
  (IN)  
  (OUT) 
[DIAGNOSTIC :]
[CALL :]
=*/
{
  cerr << "." ;
  this->ServerDBVue::freepage(pid) ;
}


/*= EXTERNAL FUNCTION =======================================================*/
void RemoteDBVue::readfirst(Page& p, const PID& pid)
/*=
[DESCRIPTION :]
  (IN)  
  (OUT) 
[DIAGNOSTIC :]
[CALL :]
=*/
{
  DBG(yoo_debug("-> readfirst", this->_vue)) ;
#ifdef DEBUG
  count_read++ ;
  if (count_read % 100 == 0) {
    cout << "Nombre de pages lues : " << count_read << endl ;
  }
#endif
  
  int err ;
  if (_deltavue == NULL) {
    if (YOO_DBServer->setLock(this->_vue, pid, YOO_SHARE) == FALSE) {
      err = YOO_ERR_LOCK ;
    } else {
      this->_locks += pid ;
      err = readdirect(&p, pid) ;
    }
  } else {
    if (this->_deltavue->read(&p, pid) == -1) {
      err = readdirect(&p, pid) ;
    } else {
      err = YOO_ERR_NONE ;
    }
  }

  if (err != YOO_ERR_NONE) {
    DBG(yoo_debug("<- readfirst error", this->_vue)) ;
  } else {
    DBG(yoo_debug("<- readfirst", this->_vue)) ;
  }
  p.setError(err) ;
}


/*= EXTERNAL FUNCTION =======================================================*/
void RemoteDBVue::readnext(Page& p, const PID& pid)
/*=
[DESCRIPTION :]
  (IN)  
  (OUT) 
[DIAGNOSTIC :]
[CALL :]
=*/
{
  DBG(yoo_debug("-> readnext", this->_vue)) ;

  int err ;
  if (_deltavue == NULL) {
    err = readdirect(&p, pid) ;
  } else {
    if (this->_deltavue->read(&p, pid) == -1) {
      err = readdirect(&p, pid) ;
    } else {
      err = YOO_ERR_NONE ;
    }
  }

  if (err != YOO_ERR_NONE) {
    DBG(yoo_debug("<- readnext error", this->_vue)) ;
  } else {
    DBG(yoo_debug("<- readnext", this->_vue)) ;
  }
} 


/*= EXTERNAL FUNCTION =======================================================*/
int RemoteDBVue::lock(const PID& pid, const Lock& l)
/*=
[DESCRIPTION :]
  (IN)  
  (OUT) 
[DIAGNOSTIC :]
[CALL :]
=*/
{
  DBG(yoo_debug("-> lock", this->_vue)) ;

  if ( (_deltavue != NULL) && (_deltavue->delta(pid)) ) {
    return YOO_ERR_UPG ;
  }
  if (YOO_DBServer->setLock(this->_vue, pid, l) == FALSE) {
    DBG(yoo_debug("<- lock error", this->_vue)) ;
    return YOO_ERR_LOCK ;
  } else {
    this->_locks += pid ;
  }

  DBG(yoo_debug("<- lock", this->_vue)) ;
  return YOO_ERR_NONE ;
} 


/*= EXTERNAL FUNCTION =======================================================*/
PID RemoteDBVue::allocation(int volNo, int num)
/*=
[DESCRIPTION :]
  (IN)  
  (OUT) 
[DIAGNOSTIC :]
[CALL :]
=*/
{
  DBG(yoo_debug("-> allocation", this->_vue)) ;
  DBG(yoo_debug("<- allocation", this->_vue)) ;
  return this->ServerDBVue::allocation(volNo, num) ;
} 
  

/*= EXTERNAL FUNCTION =======================================================*/
void lostClient (SvcSocket* sc)
/*=
[DESCRIPTION :]
  (IN)  
  (OUT) 
[DIAGNOSTIC :]
[CALL :]
=*/
{
  RemoteDBVue* vue = (RemoteDBVue*)sc ;
  yoo_debug("Connection lost : ", vue->vue()) ;
  vue->disconnect() ;
}


/*= INTERNAL FUNCTION =======================================================*/
RemoteDBVue::RemoteDBVue(void*) : SvcServer("RemoteDBVue", AF_INET)
/*=
[DESCRIPTION :]
  (IN)  
  (OUT) 
[DIAGNOSTIC :]
[CALL :]
=*/
{
  this->_vue = YOO_DBServer->registerVue(this) ;
  this->_state = YOO_IDLE ;
  this->_log = NULL ;
  this->_deltavue = NULL ;
  this->_pid = -1 ;
  this->_usid = NULL ;
}


/*= INTERNAL FUNCTION =======================================================*/
RemoteDBVue::~RemoteDBVue() 
/*=
[DESCRIPTION :]
  (IN)  
  (OUT) 
[DIAGNOSTIC :]
[CALL :]
=*/
{
  if (this->_state != YOO_IDLE) {
    this->disconnect() ;
  } 
}


/*= INTERNAL FUNCTION =======================================================*/
void RemoteDBVue::garbage(const PID& pid)
/*=
[DESCRIPTION :]
  (IN)  
  (OUT) 
[DIAGNOSTIC :]
[CALL :]
=*/
{
  int res = this->_callback.askLock(pid) ;
  if ((res) || (this->_deltavue != NULL)) {
    YOO_DBServer->unlock(this->_vue, pid) ;
  }
}


/*= EXTERNAL FUNCTION =======================================================*/
void RemoteDBVue::createLog()
/*=
[DESCRIPTION :]
  (IN)  
  (OUT) 
[DIAGNOSTIC :]
[CALL :]
=*/
{
  _log = Delta::createLog(YOO_DBServer->createLogName(_vue))  ;  
}


/*= INTERNAL FUNCTION =======================================================*/
void RemoteDBVue::MTTM_exec()
/*=
[DESCRIPTION :]
[DIAGNOSTIC :]
=*/
{
  signal(SIGINT,  ((void (*) (int))Terminate) );
  signal(SIGQUIT, ((void (*) (int))Terminate) );
  signal(SIGTERM, ((void (*) (int))Terminate) );
  signal(SIGSEGV, ((void (*) (int))Terminate) );
  SvcInit() ;

  Listener* listener = new Listener ;
  YOO_Task = new ServerTask*[YOO_NUM_TRANS+1] ;
  for (int i = 0; i < YOO_NUM_TRANS; i++) {
    RemoteDBVue* vue = new RemoteDBVue((void*)NULL) ;
    vue->async() ;
    // Pour que le serveur soit avertit de la deconnection d'un client
    vue->linkBreak(lostClient) ;
    char name[256] ;
    sprintf(name,"%s_%s", YOODA_BASE, YOO_DBServer->name()) ;
    vue->export(name) ;
    YOO_ListVues.insert(vue, -1) ;
    cout << "Export is done " <<  vue->vue() << endl ;
    YOO_Task[i] = new ServerTask(listener, vue) ;
  }
  YoodaMonitor* mon = new YoodaMonitor ;
  mon->export(YOODA_BASE) ;
  YOO_Task[i] = new ServerTask(listener, mon) ;

  taskScheduler.start() ;
}


/*= EXTERNAL FUNCTION =======================================================*/
void
YoodaMonitor::listClients(ListVal<InfoClient>& l_in, ListVal<InfoClient>& l_out)
/*=
[DESCRIPTION :]
  (IN)  
  (OUT) 
[DIAGNOSTIC :]
[CALL :]
=*/
{
  int find = 0 ;
  
  IterListPtr<RemoteDBVue> iter(YOO_ListVues) ;
  for (int i = 0; iter.more(); iter.next(), i++) {
    if (iter.get()->isConnected()) {
      IterListVal<InfoClient> iterTrans(l_in) ;
      for (; iterTrans.more(); iterTrans.next()) {
	if (iterTrans.get().pid() == iter.get()->clientpid()) {
	  iterTrans.get().addServer(YOO_DBServer->name()) ;
	  find = 1 ;
	  break ;
	}
      }
      if (find == 0) {
	l_in += InfoClient(iter.get()->clientpid(),
				iter.get()->clientname(), 
				YOO_DBServer->name()) ;
      }
      find = 0 ;	
    }
  }
  l_out.capture(l_in) ;
}


/*= EXTERNAL FUNCTION =======================================================*/
void YoodaMonitor::killClient(int pid)
/*=
[DESCRIPTION :]
  (IN)  
  (OUT) 
[DIAGNOSTIC :]
[CALL :]
=*/
{
  IterListPtr<RemoteDBVue> iter(YOO_ListVues) ;
  for (int i = 0; iter.more(); iter.next(), i++) {
    if ( (iter.get()->isConnected()) && (iter.get()->clientpid()==pid) ) {
      lostClient(YOO_ListVues[i]) ;
      YOO_Task[i]->lostConnection() ;
    }
  }
}


/*= EXTERNAL FUNCTION =======================================================*/
void YoodaMonitor::killServer()
/*=
[DESCRIPTION :]
  (IN)  
  (OUT) 
[DIAGNOSTIC :]
[CALL :]
=*/
{
  Terminate() ;
}


/*= EXTERNAL FUNCTION =======================================================*/
char* YoodaMonitor::getname()
/*=
[DESCRIPTION :]
  (IN)  
  (OUT) 
[DIAGNOSTIC :]
[CALL :]
=*/
{
  return YOO_DBServer->name() ;
}

